#include <Wire.h>
#include <SPI.h>
#include <MFRC522.h>
#include <LiquidCrystal_I2C.h>
#include <Servo.h>

// LCD I2C Address (Change to 0x3F if needed)
LiquidCrystal_I2C lcd(0x27, 16, 2);

// RFID Module Settings
#define RST_PIN 9
#define SS_PIN 10  
MFRC522 rfid(SS_PIN, RST_PIN);

// Servo Motor Pins
#define SERVO_1 5
#define SERVO_2 6
#define SERVO_3 7

Servo servo1, servo2, servo3;

// Unique RFID tags assigned to each servo (Replace with your actual tag UIDs)
byte tag1_UID[] = {0x43, 0x9B, 0x85, 0x9F};  // Tag for Servo 1
byte tag2_UID[] = {0xB3, 0x15, 0x6C, 0x9F};  // Tag for Servo 2
byte tag3_UID[] = {0x53, 0x08, 0xBE, 0x4C};  // Tag for Servo 3

void setup() {
    Serial.begin(9600);
    delay(500);

    Serial.println("Initializing System...");

    // 1️⃣ Initialize LCD FIRST
    Wire.begin();
    lcd.init();
    lcd.backlight();
    lcd.setCursor(0, 0);
    lcd.print("RFID Parking");
    lcd.setCursor(0, 1);
    lcd.print("System Ready");
    Serial.println("LCD Initialized Successfully");
    delay(2000);
    lcd.clear();

    // 2️⃣ Initialize SPI & RFID
    SPI.begin();
    rfid.PCD_Init();
    pinMode(SS_PIN, OUTPUT);
    digitalWrite(SS_PIN, HIGH); 
    Serial.println("RFID Initialized Successfully");

    // 3️⃣ Attach Servos
    servo1.attach(SERVO_1);
    servo2.attach(SERVO_2);
    servo3.attach(SERVO_3);
    Serial.println("Servos Attached Successfully");

    // 4️⃣ Reset Servos to 0 position
    servo1.write(0);
    servo2.write(0);
    servo3.write(0);
    Serial.println("System Ready!");
}

void loop() {
    digitalWrite(SS_PIN, LOW); // Enable RFID

    if (!rfid.PICC_IsNewCardPresent() || !rfid.PICC_ReadCardSerial()) {
        digitalWrite(SS_PIN, HIGH); 
        return;
    }

    digitalWrite(SS_PIN, HIGH); 
    Serial.println("RFID Tag Detected");

    // Display UID on LCD & Serial Monitor
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("RFID UID:");
    
    Serial.print("RFID UID: ");
    lcd.setCursor(0, 1);
    for (byte i = 0; i < rfid.uid.size; i++) {
        Serial.print(rfid.uid.uidByte[i] < 0x10 ? " 0" : " ");
        Serial.print(rfid.uid.uidByte[i], HEX);
        lcd.print(rfid.uid.uidByte[i] < 0x10 ? " 0" : " ");
        lcd.print(rfid.uid.uidByte[i], HEX);
    }
    Serial.println();
    delay(2000);

    processRFID();
}

// Function to check which RFID was scanned and trigger respective gate
void processRFID() {
    if (compareUID(rfid.uid.uidByte, tag1_UID)) {
        activateGate(servo1, "Gate 1");
    } 
    else if (compareUID(rfid.uid.uidByte, tag2_UID)) {
        activateGate(servo2, "Gate 2");
    } 
    else if (compareUID(rfid.uid.uidByte, tag3_UID)) {
        activateGate(servo3, "Gate 3");
    } 
    else {
        Serial.println("Unauthorized Access");
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print("Access Denied!");
        delay(2000);
        lcd.clear();
    }

    digitalWrite(SS_PIN, HIGH);
    rfid.PICC_HaltA();
}

// Function to compare scanned UID with stored UIDs
bool compareUID(byte scannedUID[], byte storedUID[]) {
    for (byte i = 0; i < 4; i++) {
        if (scannedUID[i] != storedUID[i]) {
            return false;
        }
    }
    return true;
}

// Function to open and close the gate automatically
void activateGate(Servo &servo, String gateName) {
    Serial.println(gateName + " Opening...");
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print(gateName);
    lcd.setCursor(0, 1);
    lcd.print("Open!");

    servo.write(90);  // Open gate
    delay(2000);      
    servo.write(0);   // Close gate

    Serial.println(gateName + " Closed...");
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print(gateName);
    lcd.setCursor(0, 1);
    lcd.print("Closed!");
    delay(2000);
    lcd.clear();
}
